package controller;

import javafx.application.Application;
import javafx.scene.*;
import javafx.scene.input.MouseEvent;
import javafx.scene.input.ScrollEvent;
import javafx.stage.Stage;
import model.*;
import view.*;

import java.util.ArrayList;

public class Main extends Application {

	private CustomCamera camera;
	private Group group;
	private CustomScene scene;
	private CustomMaterial material;
	private CustomLight light;
	private Custom3dModel model;
	private Desk desk;

	private Board board;
	private BoardView boardView;

	@Override
	public void start(Stage stage) {

		// init --------------------------------------------------------------

		// init - some window settings
		stage.setMaximized(true); // maximized window
		stage.setTitle("Chess3D - Tuna Acikbas & Olivier Pillods"); // title

		// init - camera
		camera = new CustomCamera(1, 1000, 70); // create a camera and some settings

		// init - scene, main container
		group = new Group(); // will contain all the elements "to show" on scene
		scene = new CustomScene(group, camera.getCamera(), SceneAntialiasing.BALANCED);

		// init - materials and shaders
		material = new CustomMaterial(); // prepare all the 3D materials (shaders)

		// init - all lights
		light = new CustomLight(); // setups different custom lights
		light.addAllToScene(group); // adds all the lights to scene

		// init - import all 3D models from .obj files
		model = new Custom3dModel();

		// init - creates a wooden desk
		desk = new Desk(material);
		desk.addAllToScene(group); // adds the desk to scene

		// init - creates a (model) list of all pieces
		board = new Board();

		// init - creates a (visual) chessboard of 8x8, and all (visual) 3d pieces
		boardView = new BoardView(model, material, board);
		boardView.addAllToScene(group); // adds the board and the pieces to scene

		// event --------------------------------------------------------------

		// event - camera rotation (mouse dragging)
		camera.initMouseControl(scene.get()); // allow user to move camera with mouse, with some constraints

		// event - camera zoom (mouse scroll)
		stage.addEventHandler(ScrollEvent.SCROLL, event -> camera.scrollCamera(event.getDeltaY()));

		// event - click on (visual) pieces to show (visually) available movements. (asked from model rules)
		for(int n = 0 ; n < 64 ; n++) {
			PieceView piece = boardView.getPiece(n);
			if(piece != null) {
				piece.getObj().addEventHandler(MouseEvent.MOUSE_CLICKED, event -> clickPiece(piece));
			}
		}

		// event - click on (visual) "colored tiles" to move the selected piece there
		for(int n = 0 ; n < 64 ; n++) {
			TileView tile = boardView.getTile(n);
			tile.getObj().addEventHandler(MouseEvent.MOUSE_CLICKED, event -> clickTile(tile));
		}

		// start --------------------------------------------------------------

		// start - links scene to application and shows application
		stage.setScene(scene.get());
		stage.show();
	}

	public static void main(String[] args) {
		launch();
	}

	public void clickPiece(PieceView clickedPiece) {

		// graveyard pieces should not do anything
		if(clickedPiece.getPosition().getX() < 0) {
			return;
		}
		// if piece is colored (aka targeted), we consider it an action to apply (tile click)
		if(clickedPiece.isTargeted()) {
			clickTile(boardView.getTile(clickedPiece.getPosition().getID()));
		} else {
			// clicking a piece will ask her available moves
			Position position = clickedPiece.getPosition();
			board.setSelected(position);
			boardView.resetColor(material);
			boardView.setColor(material, position, CustomMaterial.SELECTED_TILE);
			ArrayList<Position> moves = board.getPiece(position).getAvailableMoves(board);

			// test if kamikaze, and shows the explosion range
			if(board.isKamikaze(position)) {
				ArrayList<Position> explosions = ((Kamikaze) board.getPiece(position)).getExplosion();
				for (Position explosion : explosions) {
					boardView.setColor(material, explosion, CustomMaterial.EXPLOSION_TILE);
				}
				clickedPiece.setTargeted(material, CustomMaterial.EXPLOSION_TILE);
			}

			// shows all available moves on the tiles, with colors, each tile-color having a meaning
			for (Position move : moves) {

				if (board.isFree(move)) {
					boardView.setColor(material, move, CustomMaterial.MOVABLE_TILE);
					// test for "en passant"
					if(board.isPawn(position) || board.isNwap(position)) {
						Position enemy = new Position(move.getX(), position.getY());
						boolean isEnPassant = false;
						if(board.isEnemy(enemy, position)) {
							if (board.isPawn(enemy)) {
								if (((Pawn) board.getPiece(enemy)).isEnPassant()) {
									isEnPassant = true;
								}
							}
							if (board.isNwap(enemy)) {
								if (((Nwap) board.getPiece(enemy)).isEnPassant()) {
									isEnPassant = true;
								}
							}
						}
						if(isEnPassant) {
							boardView.setColor(material, move, CustomMaterial.SPECIAL_TILE);
							boardView.setColor(material, enemy, CustomMaterial.SPECIAL_TILE);
							boardView.getPiece(enemy.getID()).setTargeted(material, CustomMaterial.ATTACK_TILE);
						}
					}
				} else if (board.isEnemy(move, position)) {
					boardView.setColor(material, move, CustomMaterial.ATTACK_TILE);
					boardView.getPiece(move.getID()).setTargeted(material, CustomMaterial.ATTACK_TILE);
				} else {
					boardView.setColor(material, move, CustomMaterial.SPECIAL_TILE); // any special move
					boardView.getPiece(move.getID()).setTargeted(material, CustomMaterial.SPECIAL_TILE);
				}
				// PROMOTION (attacking or not)
				if ((board.isPawn(position) || board.isNwap(position)) && (move.getY() == 8 || move.getY() == 1)) {
					boardView.setColor(material, move, CustomMaterial.SPECIAL_TILE);
				}
			}
		}
	}

	public void clickTile(TileView clickedTile) {
		// --- Resets enPassant ---
		for(int j = 1 ; j <= 8 ; j++) {
			for(int i = 1 ; i <= 8 ; i++) {
				Position p = new Position(i, j);
				if (board.isPawn(p)) {
					((Pawn) board.getPiece(p)).setEnPassant(false);
				}
				if (board.isNwap(p)) {
					((Nwap) board.getPiece(p)).setEnPassant(false);
				}
			}
		}

		// --- Analyses the clicked tile type, and executes related actions ---
		int state = clickedTile.getState();
		Position selected = board.getSelected();
		Position arrival = clickedTile.getPosition();
		if(state != TileView.NORMAL && state != TileView.SELECTED) {
			// --- ALL SPECIAL MOVES ---
			if(state == TileView.SPECIAL) {
				if(board.isPawn(selected) || board.isNwap(selected)) {
					// --- PROMOTION ---
					if(arrival.getY() == 8 || arrival.getY() == 1) {
						// --- ATTACK (while doing promotion) ---
						if (!board.isFree(arrival)) {
							boardView.getPiece(arrival.getID()).resetTargeted(material); // reset color
							boardView.killPiece(arrival); // place piece on graveyard (visual)
						}
						// Promotion to Queen (by default, Queen for now)
						board.mutationOfSelectedPiece(Board.QUEEN); // (on model)
						boardView.mutationPiece(model, selected, Custom3dModel.QUEEN); // (on view)
						// --- MOVE (while doing promotion)  ---
						board.movePiece(selected, arrival); // move piece (on model)
						boardView.movePiece(selected, arrival); // move piece (on 3d view)
					}
					// --- PRISE EN PASSANT ---
					else {
						// PRE : if it's not a promotion, then it's "prise en passant", because of pawn rules
						// we don't need to test y-position
						Position enemy = new Position(arrival.getX(), selected.getY());
						boardView.getPiece(enemy.getID()).resetTargeted(material); // reset color
						boardView.killPiece(enemy); // place piece on graveyard (visual)
						board.killPiece(enemy.getID()); // because we don't move "on it", we need to kill it manually (model)
						Position enPassantArrival = new Position(arrival.getX(), selected.getY()+1);
						if(board.getSelectedPiece().getColor() == Piece.BLACK) {
							enPassantArrival.setY(selected.getY()-1);
						}
						board.movePiece(selected, enPassantArrival); // move piece (on model)
						boardView.movePiece(selected, enPassantArrival); // move piece (on 3d view)
					}
				}
				// --- CASTLING (from king) ---
				if(board.isKing(selected)) {
					// PRE (1) : the arrival piece CAN ONLY BE ROOK, because of king rules
					// PRE (2) : we don't need to test if rook !hasBeenMoved(), because of king rules
					// PRE (3) : we don't need to test if king !hasBeenMoved(), because of king rules
					int x = Board.D; // king x arrival, for west-rook
					if(arrival.getX() == Board.H) {
						x = Board.F; // king x arrival, for east-rook
					}
					Position kingArrival = new Position(x, selected.getY());
					board.movePiece(selected, kingArrival); // move king (on model)
					boardView.movePiece(selected, kingArrival); // move king (on 3d view)
					board.movePiece(arrival, selected); // move rook (on model)
					boardView.movePiece(arrival, selected); // move rook (on 3d view)
				}
				// --- CASTLING (from rook) ---
				else if(board.isRook(selected)) {
					// PRE (1) : the arrival piece CAN ONLY BE KING, because of rook rules
					// PRE (2) : we don't need to test if king !hasBeenMoved(), because of rook rules
					// PRE (3) : we don't need to test if rook !hasBeenMoved(), because of rook rules
					int x = Board.D; // king x arrival, for west-rook
					if (selected.getX() == Board.H) {
						x = Board.F; // king x arrival, for east-rook
					}
					Position kingArrival = new Position(x, arrival.getY());
					board.movePiece(arrival, kingArrival); // move king (on model)
					boardView.movePiece(arrival, kingArrival); // move king (on 3d view)
					board.movePiece(selected, arrival); // move rook (on model)
					boardView.movePiece(selected, arrival); // move rook (on 3d view)
				}
			}
			// --- EXPLOSION MOVES ---
			else if (state == TileView.EXPLOSION) {
				for(int n = 0 ; n < 64 ; n++) {
					if(boardView.getTile(n).getState() == TileView.EXPLOSION) {
						if(boardView.getPiece(n) != null) {
							boardView.getPiece(n).resetTargeted(material); // reset color
							boardView.killPiece(boardView.getPiece(n).getPosition()); // place piece on graveyard (visual)
							board.killPiece(n);
						}
					}
				}
			}
			// --- STANDARD MOVES ---
			else {
				// --- ATTACK ---
				if(state == TileView.ATTACK) {
					boardView.getPiece(arrival.getID()).resetTargeted(material); // reset color
					boardView.killPiece(arrival); // place piece on graveyard (visual)
					// --- MAGICIAN CHANGE ---
					if(board.isMagician(selected)) {
						// we can't do switch statement on ".getClass()"
						Piece enemy = board.getPiece(arrival);
						Magician magician = (Magician) board.getPiece(selected);
						int color = magician.getColor();
						if (enemy.getClass() == Pawn.class) { magician.setCopiedPiece(new Pawn(selected, color)); }
						if (enemy.getClass() == Rook.class) { magician.setCopiedPiece(new Rook(selected, color)); }
						if (enemy.getClass() == Knight.class) { magician.setCopiedPiece(new Knight(selected, color)); }
						if (enemy.getClass() == Bishop.class) { magician.setCopiedPiece(new Bishop(selected, color)); }
						if (enemy.getClass() == Queen.class) { magician.setCopiedPiece(new Queen(selected, color)); }
						if (enemy.getClass() == King.class) { magician.setCopiedPiece(new King(selected, color)); }
						if (enemy.getClass() == Nwap.class) { magician.setCopiedPiece(new Nwap(selected, color)); }
						if (enemy.getClass() == Magician.class) { magician.setCopiedPiece(((Magician) enemy).getCopiedPiece()); } // lol, stealing the stealer
						if (enemy.getClass() == Giant.class) { magician.setCopiedPiece(new Giant(selected, color)); }
						if (enemy.getClass() == Kamikaze.class) { magician.setCopiedPiece(new Kamikaze(selected, color)); }
					}
				}
				// --- EN PASSANT activation ---
				if(Math.abs(selected.getY() - arrival.getY()) == 2) {
					if(board.isPawn(selected)) {
						((Pawn) board.getPiece(selected)).setEnPassant(true);
					}
					if(board.isNwap(selected)) {
						((Nwap) board.getPiece(selected)).setEnPassant(true);
					}
				}
				// --- MOVE ---
				board.movePiece(selected, arrival); // move piece (on model)
				boardView.movePiece(selected, arrival); // move piece (on 3d view)
			}

		}
		// --- Resets board coloring, doing any action or not ---
		boardView.resetColor(material);
	}
}